#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

# Copyright (C) 2010 NAKAYA Toshiharu <nakaya@momonga-linux.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Note: 
# The initial and early version before version 0.0.17 of this file
# were under Public Domain and written by NAKAYA Toshiharu
# <nakaya@momonga-linux.org>.

require 'rpm'
require 'tmpdir'

# constants
MNPB_MOMONGA_VERSION = '7'
MNPB_NONFREE_PACKAGE_LIST = '/usr/share/momonga-nonfree-package-builder/nonfree.csv'

# error constants (naive assignments)
MNPB_ERROR_PROXY = 2
MNPB_ERROR_NOT_SUPPORT_ARCH = 3
MNPB_ERROR_COULD_NOT_CONNECT = 5
MNPB_ERROR_ARGV = 10
MNPB_ERROR_UNAVAILABLE_PACKAGE = 15
MNPB_ERROR_UNAVAILABLE_ARCH = 18
MNPB_ERROR_PACKAGE_NOTFOUND = 20
MNPB_ERROR_SVN_EXPORT = 30
MNPB_ERROR_ONLY_IX86 = 35
MNPB_ERROR_DEPENDENCY = 40
MNPB_ERROR_YUM = 50
MNPB_ERROR_SOURCES_NOTFOUND = 55
MNPB_ERROR_DOWNLOAD = 60
MNPB_ERROR_CHECKSUM = 65
MNPB_ERROR_RPMBUILD = 70

# global data
MO_SVN = Hash.new
MO_SVN['trunk'] = 'http://svn.momonga-linux.org/svn/pkgs/trunk/pkgs/'
1.upto(MNPB_MOMONGA_VERSION.to_i) { |n|
  MO_SVN[n.to_s] = "http://svn.momonga-linux.org/svn/pkgs/branches/STABLE_#{n}/pkgs/"
}
URL_ALIAS = Hash.new
URL_ALIAS[/^http:\/\/dl\.sourceforge\.net/] = 'http://jaist.dl.sourceforge.net'

# options
show_list = false
proxy_host = ''
proxy_port = ''
with_install = false
debug = false

require 'optparse'
opt = OptionParser.new
opt.on('-l', '--list', 'Show available package list') { |v| show_list = true }
opt.on('--proxy-host HOST', 'Set proxy host',String) { |v| proxy_host = v }
opt.on('--proxy-port PORT', 'Set proxy port', String) { |v| proxy_port = v }
opt.on('-i', '--install', 'Install packages after build') { |v| with_install = true }
opt.on('-d', '--debug', 'Debug mode (do not remove build directory)') { |v| debug = true }
opt.parse!(ARGV)

# make package name list
require 'csv'
nonfrees = Hash.new
CSV.foreach(MNPB_NONFREE_PACKAGE_LIST) { |r|
  name = r.shift
  nonfrees[name] = r.sort
}

# show available package list
if show_list
  title_left = 'Available package name'
  title_right = 'Available architecture'

  max_name_length = nonfrees.keys.sort { |x, y|
    y.length <=> x.length
  }.shift.length
  max_name_length = title_left.length if title_left.length > max_name_length
  column_left_width = max_name_length + 3 # 3 is mojo number

  puts ' Available package name   Available architecture'
  nonfrees.keys.sort.each { |k|
    space_width = column_left_width - k.length
    puts ' ' + k + ' ' * space_width + nonfrees[k].join(', ')
  }

  exit 0
end

# verify proxy options
if proxy_host.empty? ^ proxy_port.empty?
  STDERR.puts "Error: invalid proxy configurations"
  STDERR.puts "       I want both HOST and PORT"
  exit MNPB_ERROR_PROXY
end

# add proxy options to svn if you need
svn_exec = 'svn'
unless proxy_host.empty? || proxy_port.empty?
  svn_exec << " --config-option servers:global:http-proxy-host=#{proxy_host}"
  svn_exec << " --config-option servers:global:http-proxy-port=#{proxy_port}"
end

# this script works on ix86 and x86_64 only
require 'rbconfig'
arch = Config::CONFIG['host_cpu']
unless /^i[3456]86/ =~ arch || 'x86_64' == arch
  STDERR.puts "Error: #{$0} supports ix86 and x86_64 only"
  exit MNPB_ERROR_NOT_SUPPORT_ARCH
end

# build one package only
unless ARGV.length == 1
  STDERR.puts "Usage: #{$0} [options] package_name"
  exit MNPB_ERROR_ARGV
end

# build a package which exists in nonfree.csv only
pkgname = ARGV[0]
unless nonfrees.keys.include?(pkgname)
  STDERR.puts "Error: #{pkgname} is not available"
  exit MNPB_ERROR_UNAVAILABLE_PACKAGE
end

# check available architecture
unless nonfrees[pkgname].include?(arch)
  STDERR.puts "Error: #{pkgname} is not available on #{arch}"
  exit MNPB_ERROR_UNAVAILABLE_ARCH
end

# check momonga-release
momonga_release = `lsb_release -r`.split(/\s+/)[-1]
momonga_codename = `lsb_release -c`.split(/\s+/)[-1]
momonga_release = 'trunk' if momonga_codename == 'Sinji'
momonga_svntree = MO_SVN[momonga_release]

# check the existence of the target branch
# if false then fallback to trunk
system("#{svn_exec} ls #{momonga_svntree} > /dev/null 2>&1")
if $?.to_i.nonzero?
  STDERR.puts "Warning: I don't know release \"#{momonga_release}\""
  STDERR.puts "         So I use \"trunk\" instead of \"STABLE_#{momonga_release}\""
  momonga_svntree = MO_SVN['trunk'] 
end

# check the status of networks
system("#{svn_exec} ls http://svn.momonga-linux.org/svn/pkgs > /dev/null 2>&1")
if $?.to_i.nonzero?
  STDERR.puts "Error: #{pkgname} can not connect svn.momonga-linux.org"
  exit MNPB_ERROR_COULD_NOT_CONNECT
end

# create temporal directory
ppid = Process.ppid
pname = `ps -p #{ppid} -o args=`.split[1]
if !pname.nil? && File.basename(pname) == 'momonga-nonfree-package-builder'
  Dir.mkdir("/var/tmp/mnp#{ppid}")
  topdir = "/var/tmp/mnp#{ppid}"
else
  topdir = Dir.mktmpdir('mnp', '/var/tmp')
end

current_dir = Dir.pwd

Dir.chdir(topdir)

# check the existence of "package_name"
system("#{svn_exec} ls #{momonga_svntree}/#{pkgname} > /dev/null 2>&1")
if $?.to_i.nonzero?
  STDERR.puts "Error: #{pkgname} is not found at svn.momonga-linux.org"
  exit MNPB_ERROR_PACKAGE_NOTFOUND
end

# create rpmmacros
`echo %_topdir #{topdir} > rpmmacros`

# svn export
svn_export_spec = "#{svn_exec} export #{momonga_svntree}/#{pkgname} SOURCES"

puts "\nSVN: #{svn_export_spec}\n\n"

system(svn_export_spec)
if $?.to_i.nonzero?
  STDERR.puts "Error: some error occurred when svn checkout from svn.momonga-linux.org"
  exit MNPB_ERROR_SVN_EXPORT
end

# check the arch dependency
# packages for x86_64 only do not exist currently (2010-08-25)
if FileTest.exist?('SOURCES/NOT.x86_64') && arch == 'x86_64'
  STDERR.puts "Error: #{pkgname} works only ix86"
  exit MNPB_ERROR_ONLY_IX86
end

# move the spec file to SPECS
Dir.mkdir('SPECS')
FileUtils.mv("SOURCES/#{pkgname}.spec", 'SPECS/')

# read spec
spec = RPM::Spec.open("SPECS/#{pkgname}.spec")

# install buildreqs
def flags_to_s(flags)
  case flags
  when RPM::SENSE_LESS # 2
    '<'
  when RPM::SENSE_GREATER # 4
    '>'
  when RPM::SENSE_EQUAL # 8
    '=='
  when RPM::SENSE_LESS | RPM::SENSE_EQUAL # 10
    '<='
  when RPM::SENSE_GREATER | RPM::SENSE_EQUAL # 12
    '>='
  else # unknown
    STDERR.puts "Error: unknown dependency flag #{flags} was found"
    exit MNPB_ERROR_DEPENDENCY
  end
end

yum_runspec = 'env LANG=C sudo yum' 
unless proxy_host.empty? || proxy_port.empty?
  yum_runspec << " --setopt=proxy=http://#{proxy_host}:#{proxy_port}/"
end
yum_runspec << ' -y install build-momonga'

spec.buildrequires.each { |dep|
  if dep.flags.zero?
    nevr = dep.name
  else
    nevr = "'#{dep.name} #{flags_to_s(dep.flags)} #{dep.version.to_vre}'"
  end
  yum_runspec << " #{nevr}"
}

puts "\nYUM: #{yum_runspec}\n\n"

system(yum_runspec)
if $?.to_i.nonzero?
  STDERR.puts "Error: installation of buildreqs through yum was failed"
  exit MNPB_ERROR_YUM
end

# download NoSource files
Dir.chdir('SOURCES')

# make NoSource file list
to_downlaod = []
spec.sources.reverse.each { |src|
  to_downlaod << src.fullname if src.no?
}

# check the existence of sources file if NoSource
unless to_downlaod.empty? || FileTest.exist?('sources')
  STDERR.puts "Error: sources file does not found"
  exit MNPB_ERROR_SOURCES_NOTFOUND
end

# download using wget
status = 0
to_downlaod.each { |nosrc|
  re = nil
  URL_ALIAS.each_key { |key|
    if key.match(nosrc)
      re = key
      break
    end
  }
  nosrc.sub!(re, URL_ALIAS[re]) if re

  wget_runspec = 'wget --inet4-only'
  unless proxy_host.empty? || proxy_port.empty?
    wget_runspec << " --execute=http_proxy=#{proxy_host}:#{proxy_port}"
  end
  wget_runspec << " #{nosrc}"

  puts "\nWGET: #{wget_runspec}\n\n"

  system(wget_runspec)
  status |= $?.to_i
}

if status.nonzero?
  STDERR.puts "Error: some error occurred when download"
  exit MNPB_ERROR_DOWNLOAD
end

# calculate checksums
checksum_result = nil
checksums = Hash.new

if File.exist?('sources')
  open('sources') { |f|
    while l = f.gets
      if /^([0-9abcdef]+)\s+([^\s]+)/ =~ l
        checksums[File.basename($2)] = $1
      end
    end
  }

  checksum_result = to_downlaod.map { |nosrc|
    filename = File.basename(nosrc)
    if checksums.has_key?(filename)
      sum = `sha256sum #{filename}`.split[0]
      cmp = (sum == checksums[filename])
      unless cmp
        STDERR.puts "Error: checksum of #{filename} does not match"
        STDERR.puts "       #{sum}: now downloaded file"
        STDERR.puts "       #{checksums[filename]}: written on the sources file"
      end
      cmp
    else
      STDERR.puts "Error: #{filename} has no checksum"
      false
    end
  }

  unless checksum_result.inject(true) { |x, y| x && y }
    exit MNPB_ERROR_CHECKSUM
  end
end

Dir.chdir('..')

# rpmbuild
rpmbuild_runspec = 
  'rpmbuild -ba ' \
  '--rcfile /usr/lib/rpm/rpmrc:' \
  '/usr/lib/rpm/momonga/rpmrc' \
  ' ' \
  '--macros /usr/lib/rpm/macros:' \
  '/usr/lib/rpm/platform/%{_target}/macros:' \
  '/usr/lib/rpm/momonga/macros:' \
  '/etc/rpm/macros.*:' \
  '/etc/rpm/macros:' \
  '/etc/rpm/%{_target}/macros:' \
  '~/.rpmmacros:' \
  './rpmmacros' \
  ' ' \
  "SPECS/#{pkgname}.spec"

# XXX: workaround for mo6 or earlier (this rpmrc is broken)
unless FileTest.exists?('/usr/lib/rpm/momonga/rpmrc')
  File.open('./rpmrc', 'w') { |f|
    f.print <<EOF
buildarchtranslate: geode: i586
buildarchtranslate: athlon-mp: athlon
buildarchtranslate: athlon-xp: athlon
buildarchtranslate: athlon-4: athlon
buildarchtranslate: athlon-tbird: athlon
buildarchtranslate: athlon: i686
buildarchtranslate: k6-3: k6
buildarchtranslate: k6-2: k6
buildarchtranslate: k6: i586
buildarchtranslate: pentium4: i686
buildarchtranslate: pentium3: i686
buildarchtranslate: pentium2: i686
buildarchtranslate: pentium-mmx: i586
buildarchtranslate: i686: i686
buildarchtranslate: i586: i586
buildarchtranslate: i486: i486
buildarchtranslate: i386: i386

optflags: i686 -O2 -mtune=generic -march=i686 -fstack-protector
optflags: x86_64 -O2 -mtune=generic -fPIC -fstack-protector
EOF
  }

  rpmbuild_runspec = 
    'rpmbuild -ba ' \
    '--rcfile /usr/lib/rpm/rpmrc:' \
    './rpmrc' \
    ' ' \
    '--macros /usr/lib/rpm/macros:' \
    '/usr/lib/rpm/platform/%{_target}/macros:' \
    '/usr/lib/rpm/momonga/macros:' \
    '/etc/rpm/macros.*:' \
    '/etc/rpm/macros:' \
    '/etc/rpm/%{_target}/macros:' \
    '~/.rpmmacros:' \
    './rpmmacros' \
    ' ' \
    "SPECS/#{pkgname}.spec"
end

puts "\nRPMBUILD: #{rpmbuild_runspec}\n\n"

system(rpmbuild_runspec)
if $?.to_i.nonzero?
  STDERR.puts "Error: rpmbuild failure occurred"
  exit MNPB_ERROR_RPMBUILD
end

# move *.rpm to current_dir
rpmlst = []
Dir.glob("RPMS/**/*.rpm").each { |rpm|
  filename = File.basename(rpm)
  FileUtils.copy_file(rpm, "#{File.expand_path(current_dir)}/#{filename}", true)
  rpmlst << filename
}

# install the binary packages
if with_install
  pkgs = []
  Dir.glob("RPMS/**/*.rpm").each { |rpm|
    n = File.basename(rpm).split('-')[0..-3].join('-')
    pkgs << n
  }
  if pkgs.include?(pkgname)
    system('createrepo -v -d RPMS')
    File.open('momonga-nonfree-package-builder.repo', 'w') { |f|
      f.print <<EOF
[momonga-nonfree-package-builder]
name=Momonga Nonfree Package Builder
baseurl=file://#{topdir}/RPMS
enabled=1
gpgcheck=0
EOF
    }

    yum_runspec = 'env LANG=C sudo yum -c momonga-nonfree-package-builder.repo'
    unless proxy_host.empty? || proxy_port.empty?
      yum_runspec << " --setopt=proxy=http://#{proxy_host}:#{proxy_port}/"
    end
    yum_runspec << " -y install #{pkgname}"

    puts "\nYUM: #{yum_runspec}\n\n"

    system(yum_runspec)
    if $?.to_i.nonzero?
      STDERR.puts "Error: installation of packages through yum was failed"
      exit MNPB_ERROR_INSTALL
    end
  else
    STDERR.puts "Warning: the package which has the name '#{pkgname}' is not found"
    STDERR.puts "         Please install manually"
  end
end

# remove build directory
system("rm -rf #{topdir}") unless debug

# show success message
Dir.chdir(current_dir)
rpmlst.sort!
is_first = true
rpmlst.each { |rpm|
  if is_first
    print "\nDone: "
    is_first = false
  else
    print '      '
  end
  puts rpm
}
